#include<linux/module.h>
#include<linux/kthread.h>
#include<linux/wait.h>
#include<linux/seqlock.h>

enum thread_index {WAKING_THREAD, WRITER_THREAD, FIRST_READER_THREAD, SECOND_READER_THREAD};

static struct thread_structure 
{
	struct task_struct *thread[4];
} threads;

static wait_queue_head_t wait_queue;
static bool condition;
static int number;
static DEFINE_SEQLOCK(number_lock);
static const int first_reader_number = 1, second_thread_number = 2;

static int reader_thread(void *data) 
{
	DEFINE_WAIT(wait);
	unsigned long int seqlock_value = 0;
	int local_number = 0;
	for(;;) {
		do {
				seqlock_value = read_seqbegin(&number_lock);
				local_number = number;
		} while(read_seqretry(&number_lock, seqlock_value));
		pr_info("[reader_number: %d] Value of \"number\" variable: %d\n", *(int *)data,local_number);
		add_wait_queue(&wait_queue,&wait);
		while(!condition) {
			prepare_to_wait(&wait_queue,&wait,TASK_INTERRUPTIBLE); 
			if(kthread_should_stop()) 
				return 0;
			pr_info("[reader_thread %d]: awake\n", *(int *)data);
			schedule();
		}
		condition=false;
		finish_wait(&wait_queue,&wait); 
	}
}

static int writer_thread(void *data) 
{
	for(;;) {
		write_seqlock(&number_lock);
		number++;
		write_sequnlock(&number_lock);
		if(kthread_should_stop()) 
			return 0;
		set_current_state(TASK_INTERRUPTIBLE);
		if(schedule_timeout(HZ>>2))
			pr_info("Signal received!\n");
		pr_info("[writer_thread] awake!\n");
	}
}

static int waking_thread(void *data)
{
	for(;;) {
		if(kthread_should_stop())
			return 0;
		set_current_state(TASK_INTERRUPTIBLE);
		if(schedule_timeout(1*HZ))
			pr_info("Signal received!\n");
		condition=true;
		wake_up_all(&wait_queue);
	}

}

static int __init threads_init(void)
{
	init_waitqueue_head(&wait_queue);
	threads.thread[WRITER_THREAD] = kthread_run(writer_thread,NULL,"writer_thread");
	threads.thread[FIRST_READER_THREAD] = kthread_run(reader_thread,(void *)&first_reader_number,"first_reader_thread");
	threads.thread[SECOND_READER_THREAD] = kthread_run(reader_thread,(void *)&second_thread_number,"second_reader_thread");
	threads.thread[WAKING_THREAD] = kthread_run(waking_thread,NULL,"waking_thread");
	return 0;
}

static void __exit threads_exit(void)
{
	kthread_stop(threads.thread[WAKING_THREAD]);
	kthread_stop(threads.thread[WRITER_THREAD]);
	kthread_stop(threads.thread[FIRST_READER_THREAD]);
	kthread_stop(threads.thread[SECOND_READER_THREAD]);
}

module_init(threads_init);
module_exit(threads_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("An example of using the Linux kernel threads and a seqential lock.");
MODULE_AUTHOR("Arkadiusz Chrobot <a.chrobot@tu.kielce.pl>");
